"""
Useful utilities for ensemble
"""

from typing import Any, List, Optional
import tensorflow as tf
import numpy as np

NDArray = Any
kwargus = Any
model_type = Any


class bagging:  # A.K.A. voting
    def __init__(self) -> None:
        self.models: List[model_type] = []
        self.model_trained: List[bool] = []
        self.count = 0
        self.need_confidence = True  # Help in reducing numbers of get_confidence runs
        self.permit_train = False

    def append(self, model: model_type, model_trained: bool) -> None:
        """
        Add model to the voting method
        """
        self.models.append(model)
        self.model_trained.append(model_trained)
        self.count += 1

    def __train_model(self, i: int, **kwargs: kwargus) -> None:
        """
        Train a model if it isn't trained already
        """
        if not self.model_trained[i]:
            self.need_confidence = True
            self.model_trained[i] = True
            self.models[i].fit(**kwargs)

    def train(self, **kwargs: kwargus) -> None:
        """
        Train all models in the class, **kwargs expect to receive the argus that can be directly sent to tf.fit
        Expected to be run after finishing compile
        """
        if not self.permit_train:
            raise ValueError("Models needed to be compiled before training")
        for i in range(self.count):
            if "verbose" in kwargs:
                if kwargs["verbose"] == 1:
                    print("Model ", i + 1, "/", self.count, " is training...")
            else:
                print("Model ", i + 1, "/", self.count, " is training...")
            self.__train_model(i, **kwargs)

    def compile(self, **kwargs: kwargus) -> None:
        self.permit_train = True

        for i in range(self.count):
            if not self.model_trained[i]:
                dict_kwargs = kwargs.copy()
                # TODO(@refraction-ray): still not compatible with new optimizer
                # https://github.com/tensorflow/tensorflow/issues/58973
                self.models[i].compile(**dict_kwargs)

    def __get_confidence(self, model_index: int, input: NDArray) -> NDArray:
        """
        Get the confidence value that is needed by voting.
        Number of calling this function is reduced by self.need_confidence
        """
        self.need_confidence = False
        prediction = self.models[model_index].predict(input, verbose=0)
        prediction_returns = np.zeros(len(prediction))
        for i in range(len(prediction)):
            prediction_returns[i] = prediction[i][0]
        return prediction_returns

    """
    Voting strategies begin
    More voting strategies can be added beneath, a single function, and a if function in self.predict
    """

    def __voting_weight(self, array: NDArray) -> NDArray:
        result = []
        for i in array:
            result.append(self.__voting_weight_single(i))
        return np.array(result)

    def __voting_average(self, array: NDArray) -> NDArray:
        result = np.mean(array, axis=1)
        return result

    def __voting_weight_single(self, array: NDArray) -> float:
        opp_array = np.ones(len(array)) - array
        weight = np.absolute(opp_array - array)
        weight_sum = np.sum(weight)
        weight = weight / weight_sum
        result = array * weight
        return float(np.sum(result))

    def __voting_most(self, array: NDArray) -> NDArray:
        return_result = []
        for items in array:
            items_ = items > 0.5
            result = 0
            for i in items_:
                result += 1 if i else -1
            return_result.append(1 if result > 0 else 0)
        return np.array(return_result)

    def predict(
        self, input_data: NDArray, voting_policy: Optional[str] = None
    ) -> NDArray:
        """
        Input data is expected to be a 2D array that the first layer is different input data (into the trained models)
        """
        if self.need_confidence:
            predictions = []
            for i in range(self.count):
                predictions.append(np.array(self.__get_confidence(i, input_data)))
            self.predictions = np.transpose(np.array(predictions))
        if voting_policy == "weight":
            return self.__voting_weight(self.predictions)
        elif voting_policy == "most":
            return self.__voting_most(self.predictions)
        elif voting_policy == "average":
            return self.__voting_average(self.predictions)
        elif voting_policy is None:
            return self.predictions
        else:
            raise ValueError("voting_policy must be none, weight, most, or average")

    def __acc_binarify(self, array: NDArray) -> NDArray:
        """
        Needed for ACC test
        """
        result = []
        for i in array:
            result.append(1 if (i > 0.5) else 0)
        return result

    def __eval_accuracy(self, input_data: NDArray) -> float:
        input_data[1] = self.__acc_binarify(input_data[1])
        algo = tf.keras.metrics.Accuracy()
        algo.reset_state()
        algo.update_state(input_data[0], input_data[1])
        return float(algo.result().numpy())

    def __eval_auc(self, input_data: NDArray) -> float:
        algo = tf.keras.metrics.AUC()
        algo.reset_state()
        algo.update_state(input_data[0], input_data[1])
        return float(algo.result().numpy())

    def eval(self, input_data: List[NDArray], evaluation_method: str = "acc") -> float:
        """
        Expect input data to be a 2D array
        which a 1D array of yTrue followed by a 1D array of yPred is expected to be the components of the 2D array
        """
        if evaluation_method == "acc":
            return self.__eval_accuracy(input_data)
        elif evaluation_method == "auc":
            return self.__eval_auc(input_data)
        else:
            raise ValueError("evaluation_method must be acc or auc")



"""
One-hot variational autoregressive models for multiple categorical choices beyond binary
"""

from typing import Any, Optional, Tuple, Union, List
import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras.models import Model
import numpy as np


class MaskedLinear(Layer):  # type: ignore
    def __init__(
        self,
        input_space: int,
        output_space: int,
        spin_channel: int,
        mask: Optional[tf.Tensor] = None,
        dtype: Optional[tf.DType] = None,
    ):
        super().__init__()
        if dtype is None:
            self._dtype = tf.float32
        else:
            self._dtype = dtype
        self.w = self.add_weight(
            name="w",
            shape=(output_space, spin_channel, input_space, spin_channel),
            initializer="random_normal",
            trainable=True,
            dtype=self._dtype,
        )
        self.b = self.add_weight(
            name="b",  # every variables in customized layer must have a explicit name so that the model can be saved
            # see https://github.com/tensorflow/tensorflow/issues/26811
            shape=(output_space, spin_channel),
            initializer="random_normal",  # zero is not ok, as the first output will always equal probabilitied
            trainable=True,
            dtype=self._dtype,
        )
        self.input_space = input_space
        self.output_space = output_space
        self.spin_channel = spin_channel
        if mask is not None:
            self.mask = mask
        else:
            self.mask = tf.ones([output_space, spin_channel, input_space, spin_channel])

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        w_mask = tf.multiply(self.mask, self.w)
        return tf.tensordot(inputs, w_mask, axes=[[-2, -1], [2, 3]]) + self.b

    def regularization(self, lbd_w: float = 1.0, lbd_b: float = 1.0) -> tf.Tensor:
        return lbd_w * tf.reduce_sum(self.w**2) + lbd_b * tf.reduce_sum(self.b**2)


class MADE(Model):  # type: ignore
    def __init__(
        self,
        input_space: int,
        output_space: int,
        hidden_space: int,
        spin_channel: int,
        depth: int,
        evenly: bool = True,
        dtype: Optional[tf.DType] = None,
        activation: Optional[Any] = None,
        nonmerge: bool = True,
        probamp: Optional[tf.Tensor] = None,
    ):
        super().__init__()
        if not dtype:
            self._dtype = tf.float32
        else:
            self._dtype = dtype
        self._m: List[np.array] = []  # type: ignore
        self._masks: List[tf.Tensor] = []
        self.ml_layer: List[Any] = []
        self.input_space = input_space
        self.spin_channel = spin_channel
        self._build_masks(
            input_space, output_space, hidden_space, spin_channel, depth, evenly
        )
        self._check_masks()
        self.depth = depth
        self.relu = []
        if activation is None:
            activation = tf.keras.layers.PReLU
        for _ in range(depth):
            self.relu.append(activation())
        self.nonmerge = nonmerge
        self.probamp = probamp

    def _build_masks(
        self,
        input_space: int,
        output_space: int,
        hidden_space: int,
        spin_channel: int,
        depth: int,
        evenly: bool,
    ) -> None:
        self._m.append(np.arange(1, input_space + 1))
        for i in range(1, depth + 1):
            if i == depth:  # the last layer
                # assign output layer units a number between 1 and D
                m = np.arange(1, input_space + 1)
                assert (
                    output_space % input_space == 0
                ), "output space must by multiple of input space"
                self._m.append(
                    np.hstack([m for _ in range(output_space // input_space)])
                )
            else:  # middle layer
                # assign hidden layer units a number between 1 and D-1
                if evenly:
                    assert (
                        hidden_space % (input_space - 1) == 0
                    ), "hidden space must be multiple of input space -1 when you set evenly as True"
                    m = np.arange(1, input_space)
                    # TODO: whether 0 is ok need further scrunity, this is a bit away from original MADE idea
                    self._m.append(
                        np.hstack([m for _ in range(hidden_space // (input_space - 1))])
                    )
                else:
                    self._m.append(np.random.randint(1, input_space, size=hidden_space))
            if i == depth:
                mask = self._m[i][None, :] > self._m[i - 1][:, None]  # type: ignore
            else:
                # input to hidden & hidden to hidden
                mask = self._m[i][None, :] >= self._m[i - 1][:, None]  # type: ignore
            # add two spin channel axis into mask
            mask = mask.T
            mask = mask[:, np.newaxis, :, np.newaxis]
            mask = np.tile(mask, [1, spin_channel, 1, spin_channel])
            self._masks.append(tf.constant(mask, dtype=self._dtype))
            if i == depth and depth == 1:
                # in case there is only one layer
                self.ml_layer.append(
                    MaskedLinear(
                        input_space,
                        output_space,
                        spin_channel,
                        mask=self._masks[i - 1],
                        dtype=self._dtype,
                    )
                )
            elif i == depth:
                self.ml_layer.append(
                    MaskedLinear(
                        hidden_space,
                        output_space,
                        spin_channel,
                        mask=self._masks[i - 1],
                        dtype=self._dtype,
                    )
                )
            elif i == 1:
                self.ml_layer.append(
                    MaskedLinear(
                        input_space,
                        hidden_space,
                        spin_channel,
                        mask=self._masks[i - 1],
                        dtype=self._dtype,
                    )
                )
            else:
                self.ml_layer.append(
                    MaskedLinear(
                        hidden_space,
                        hidden_space,
                        spin_channel,
                        mask=self._masks[i - 1],
                        dtype=self._dtype,
                    )
                )

    def _check_masks(self) -> None:
        pass

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        if self.nonmerge:
            nonmerge_block = tf.roll(inputs, axis=-2, shift=1) * -10.0
            # PBC here
            # once the last pixel is i, the next one cannot be i
        for i in range(self.depth):
            inputs = self.ml_layer[i](inputs)
            inputs = self.relu[i](inputs)
        if self.probamp is not None:
            inputs = self.probamp + inputs
        if self.nonmerge:
            inputs = nonmerge_block + inputs
            # TODO: to be investigated whether the nonmerge block can affect on unity of probability
        outputs = tf.keras.layers.Softmax()(inputs)
        return outputs

    def model(self) -> Any:
        x = tf.keras.layers.Input(shape=(self.input_space, self.spin_channel))
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

    def regularization(self, lbd_w: float = 1.0, lbd_b: float = 1.0) -> tf.Tensor:
        loss = 0.0
        for l in self.ml_layer:
            loss += l.regularization(lbd_w=lbd_w, lbd_b=lbd_b)
        return loss

    def sample(self, batch_size: int) -> Tuple[np.array, tf.Tensor]:  # type: ignore
        eps = 1e-10
        sample = np.zeros(
            [batch_size, self.input_space, self.spin_channel], dtype=np.float32
        )
        for i in range(self.input_space):
            x_hat = self.call(sample)
            ln_x_hat = tf.math.log(
                tf.reshape(x_hat, [batch_size * self.input_space, self.spin_channel])
                + eps
            )
            choice = tf.random.categorical(ln_x_hat, num_samples=1)
            choice = tf.reshape(choice, [choice.shape[0]])
            one_hot = tf.one_hot(choice, depth=self.spin_channel)
            one_hot = tf.reshape(
                one_hot, [batch_size, self.input_space, self.spin_channel]
            )
            sample[:, i, :] = one_hot[:, i, :]

        return sample, x_hat

    def _log_prob(self, sample: tf.Tensor, x_hat: tf.Tensor) -> tf.Tensor:
        eps = 1e-10
        probm = tf.multiply(x_hat, sample)
        probm = tf.reduce_sum(probm, axis=-1)
        lnprobm = tf.math.log(probm + eps)
        return tf.reduce_sum(lnprobm, axis=-1)

    def log_prob(self, sample: tf.Tensor) -> tf.Tensor:
        x_hat = self.call(sample)
        log_prob = self._log_prob(sample, x_hat)
        return log_prob


class MaskedConv2D(Layer):  # type: ignore
    def __init__(self, mask_type: str, **kwargs: Any):
        super().__init__()
        assert mask_type in {"A", "B"}, "mask_type must be in A or B"
        self.mask_type = mask_type
        self.conv = tf.keras.layers.Conv2D(**kwargs)

    def build(self, input_shape: Union[List[tf.TensorShape], tf.TensorShape]) -> None:
        # Build the conv2d layer to initialize kernel variables
        self.conv.build(input_shape)
        # Use the initialized kernel to create the mask
        kernel_shape = self.conv.kernel.shape
        self.kernel = self.add_weight(shape=self.conv.kernel.shape)
        self.bias = self.add_weight(shape=self.conv.bias.shape)
        self.mask = np.zeros(shape=kernel_shape)
        self.mask[: kernel_shape[0] // 2, ...] = 1.0
        self.mask[kernel_shape[0] // 2, : kernel_shape[1] // 2, ...] = 1.0
        if self.mask_type == "B":
            self.mask[kernel_shape[0] // 2, kernel_shape[1] // 2, ...] = 1.0

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        # self.conv.kernel = self.kernel * self.mask
        # self.conv.bias = self.bias
        self.conv.set_weights([self.kernel * self.mask, self.bias])
        return self.conv(inputs)


class ResidualBlock(Layer):  # type: ignore
    def __init__(self, layers: List[Any]):
        super().__init__()
        self.layers = layers

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        y = inputs
        for layer in self.layers:
            y = layer(y)
        return y + inputs


class PixelCNN(Model):  # type: ignore
    def __init__(self, spin_channel: int, depth: int, filters: int):
        super().__init__()
        self.rb = []
        self.rb0 = MaskedConv2D(
            mask_type="A", filters=filters, kernel_size=3, padding="same"
        )

        self.relu = tf.keras.layers.PReLU()
        for _ in range(1, depth):
            self.rb.append(
                ResidualBlock(
                    [
                        MaskedConv2D(
                            mask_type="B",
                            filters=filters,
                            kernel_size=3,
                            padding="same",
                        ),
                        tf.keras.layers.PReLU(),
                    ]
                )
            )
        self.rbn = MaskedConv2D(
            mask_type="B", filters=spin_channel, kernel_size=1, padding="same"
        )
        self.depth = depth
        self.spin_channel = spin_channel

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        x = self.rb0(inputs)
        x = self.relu(x)
        for i in range(self.depth - 1):
            x = self.rb[i](x)
        x = self.rbn(x)
        x = tf.keras.layers.Softmax(axis=-1)(x)
        return x

    def sample(self, batch_size: int, h: int, w: int) -> Tuple[np.array, tf.Tensor]:  # type: ignore
        eps = 1e-10
        sample = np.zeros([batch_size, h, w, self.spin_channel], dtype=np.float32)
        for i in range(h):
            for j in range(w):
                x_hat = self.call(sample)
                ln_x_hat = tf.math.log(
                    tf.reshape(x_hat, [batch_size * h * w, self.spin_channel]) + eps
                )
                choice = tf.random.categorical(ln_x_hat, num_samples=1)
                choice = tf.reshape(choice, [choice.shape[0]])
                one_hot = tf.one_hot(choice, depth=self.spin_channel)
                one_hot = tf.reshape(one_hot, [batch_size, h, w, self.spin_channel])
                sample[:, i, j, :] = one_hot[:, i, j, :]

        return sample, x_hat

    def _log_prob(self, sample: tf.Tensor, x_hat: tf.Tensor) -> tf.Tensor:
        eps = 1e-10
        probm = tf.multiply(x_hat, sample)
        probm = tf.reduce_sum(probm, axis=-1)
        lnprobm = tf.math.log(probm + eps)
        return tf.reduce_sum(lnprobm, axis=[-1, -2])

    def log_prob(self, sample: tf.Tensor) -> tf.Tensor:
        x_hat = self.call(sample)
        log_prob = self._log_prob(sample, x_hat)
        return log_prob


class NMF(Model):  # type: ignore
    def __init__(
        self,
        spin_channel: int,
        *dimensions: int,
        _dtype: tf.DType = tf.float32,
        probamp: Optional[tf.Tensor] = None,
    ):
        super().__init__()
        self.w = self.add_weight(
            name="meanfield-parameter",
            shape=list(dimensions) + [spin_channel],
            initializer="random_normal",
            trainable=True,
            dtype=_dtype,
        )
        self._dtype = _dtype
        self.dimensions = list(dimensions)
        self.D = len(self.dimensions)
        self.spin_channel = spin_channel
        self.probamp = probamp

    def call(self, inputs: Optional[tf.Tensor] = None) -> tf.Tensor:
        # x = tf.keras.layers.Softmax(axis=-1)(self.w)
        if self.probamp is None:
            return self.w
        else:
            return self.w + self.probamp

    def sample(self, batch_size: int) -> Tuple[tf.Tensor, tf.Tensor]:
        x_hat = self.call()
        x_hat = x_hat[tf.newaxis, :]
        tile_shape = tuple([batch_size] + [1 for _ in range(self.D + 1)])
        x_hat = tf.tile(x_hat, tile_shape)
        totalsize = tf.reduce_prod(self.dimensions)
        logits = tf.reshape(x_hat, (batch_size * totalsize, self.spin_channel))
        sample = tf.random.categorical(logits, 1)  # [system size, batch size]
        # [ * dimensions, batch size, spin channel]
        sample = tf.one_hot(sample, depth=self.spin_channel)
        sample = tf.reshape(
            sample, [batch_size] + self.dimensions + [self.spin_channel]
        )
        x_hat = tf.keras.layers.Softmax(axis=-1)(x_hat)
        return sample, x_hat

    def _log_prob(self, sample: tf.Tensor, x_hat: tf.Tensor) -> tf.Tensor:
        eps = 1e-10
        probm = tf.multiply(x_hat, sample)
        probm = tf.reduce_sum(probm, axis=-1)
        lnprobm = tf.math.log(probm + eps)
        return tf.reduce_sum(lnprobm, axis=[-i - 1 for i in range(self.D)])

    def log_prob(self, sample: tf.Tensor) -> tf.Tensor:
        x_hat = self.call(sample)
        log_prob = self._log_prob(sample, x_hat)
        return log_prob



"""
Cloud provider from QuaFu: http://quafu.baqis.ac.cn/
"""

from typing import Any, Dict, List, Optional, Sequence, Union
import logging

from quafu import User, QuantumCircuit
from quafu import Task as Task_

from .abstraction import Device, sep, Task
from ..abstractcircuit import AbstractCircuit
from ..utils import is_sequence

logger = logging.getLogger(__name__)


def list_devices(token: Optional[str] = None, **kws: Any) -> List[Device]:
    raise NotImplementedError


def list_properties(device: Device, token: Optional[str] = None) -> Dict[str, Any]:
    raise NotImplementedError


def submit_task(
    device: Device,
    token: str,
    shots: Union[int, Sequence[int]] = 1024,
    circuit: Optional[Union[AbstractCircuit, Sequence[AbstractCircuit]]] = None,
    source: Optional[Union[str, Sequence[str]]] = None,
    compile: bool = True,
    **kws: Any,
) -> Task:
    if source is None:

        def c2qasm(c: Any) -> str:
            from qiskit.circuit import QuantumCircuit

            if isinstance(c, QuantumCircuit):
                s = c.qasm()
                # nq = c.num_qubits
            else:
                s = c.to_openqasm()
            return s  # type: ignore

        if not is_sequence(circuit):
            source = c2qasm(circuit)
        else:
            source = [c2qasm(c) for c in circuit]  # type: ignore
    user = User(token)
    user.save_apitoken()

    def c2task(source: str) -> Task:
        nq = int(source.split("\n")[2].split("[")[1].split("]")[0])  # type: ignore
        qc = QuantumCircuit(nq)
        qc.from_openqasm(source)
        task = Task_()
        device_name = device.name.split(sep)[-1]
        task.config(backend=device_name, shots=shots, compile=compile)
        res = task.send(qc, wait=False)
        wrapper = Task(res.taskid, device=device)
        return wrapper

    if not is_sequence(source):
        return c2task(source)  # type: ignore
    else:
        return [c2task(s) for s in source]  # type: ignore


def resubmit_task(task: Task, token: str) -> Task:
    raise NotImplementedError


def remove_task(task: Task, token: str) -> Any:
    raise NotImplementedError


def list_tasks(device: Device, token: str, **filter_kws: Any) -> List[Task]:
    raise NotImplementedError


def get_task_details(
    task: Task, device: Device, token: str, prettify: bool
) -> Dict[str, Any]:
    # id results
    r = {}
    r["id"] = task.id_
    t = Task_()
    r["results"] = dict(t.retrieve(task.id_).counts)  # type: ignore
    if r["results"]:
        r["state"] = "completed"
    else:
        r["state"] = "pending"
    return r



"""
Analog-Digital Hybrid Circuit class wrapper
only support jax backend
"""

from typing import Any, List, Optional, Callable, Dict, Tuple, Union, Sequence
from dataclasses import dataclass
from functools import partial

import numpy as np
import tensornetwork as tn

from .cons import backend, rdtypestr
from .abstractcircuit import defined_gates
from .circuit import Circuit
from .quantum import QuOperator
from .timeevol import ode_evol_global, ode_evol_local
from .utils import arg_alias

Tensor = Any


@dataclass
class AnalogBlock:
    """
    A data structure to hold information about an analog evolution block.
    """

    hamiltonian_func: Callable[[Tensor], Tensor]
    time: float
    index: Optional[List[int]] = None
    solver_options: Optional[Dict[str, Any]] = None


class AnalogCircuit:
    """
    A class for hybrid digital-analog quantum simulation with time-dependent Hamiltonians.
    """

    def __init__(
        self,
        nqubits: int,
        inputs: Optional[Tensor] = None,
        mps_inputs: Optional[QuOperator] = None,
        split: Optional[Dict[str, Any]] = None,
        dim: Optional[int] = None,
    ):
        """
        Initializes the hybrid circuit.

        :param nqubits: The number of qubits in the circuit.
        :type nqubits: int
        :param dim: The local Hilbert space dimension per site. Qudit is supported for 2 <= d <= 36.
        :type dim: If None, the dimension of the circuit will be `2`, which is a qubit system.
        :param inputs: If not None, the initial state of the circuit is taken as ``inputs``
            instead of :math:`\vert 0 \rangle^n` qubits, defaults to None.
        :type inputs: Optional[Tensor], optional
        :param mps_inputs: QuVector for a MPS like initial wavefunction.
        :type mps_inputs: Optional[QuOperator]
        :param split: dict if two qubit gate is ready for split, including parameters for at least one of
            ``max_singular_values`` and ``max_truncation_err``.
        :type split: Optional[Dict[str, Any]]
        """
        self.num_qubits, self._nqubits = nqubits, nqubits
        self.dim = 2**self.num_qubits
        self.inputs = inputs
        if inputs is None:
            self.inputs = np.zeros([self.dim])
            self.inputs[0] = 1.0
            self.inputs = backend.convert_to_tensor(self.inputs)

        # List of digital circuits, starting with one empty circuit.
        self.digital_circuits: List[Circuit] = [
            Circuit(self.num_qubits, inputs, mps_inputs, split, dim)
        ]

        # List of analog blocks, each containing the Hamiltonian function, time, and solver options.
        self.analog_blocks: List[AnalogBlock] = []
        self._effective_circuit: Optional[Circuit] = None
        self._solver_options: Dict[str, Any] = {}

    def set_solver_options(self, **kws: Any) -> None:
        """
        set solver options globally for this circuit object
        """
        self._solver_options = kws

    @property
    def effective_circuit(self) -> Circuit:
        """
        Returns the effective circuit after all blocks have been added.
        """
        if self._effective_circuit is None:
            self.state()
        return self._effective_circuit  # type: ignore

    @property
    def current_digital_circuit(self) -> Circuit:
        """
        Returns the last (currently active) digital circuit.
        """
        return self.digital_circuits[-1]

    def add_analog_block(
        self,
        hamiltonian: Callable[[float], Tensor],
        time: Union[float, List[Tensor]],
        index: Optional[List[int]] = None,
        **solver_options: Any,
    ) -> "AnalogCircuit":
        """
        Adds a time-dependent analog evolution block to the circuit.

        This finalizes the current digital block and prepares a new one for subsequent gates.

        :param hamiltonian_func: A function H(t) that takes a time `t` (from 0 to `time`)
                                 and returns the Hamiltonian matrix at that instant.
        :type hamiltonian_func: Callable[[float], np.ndarray]
        :param time: The total evolution time 'T'.
        :type time: float
        :param index: The indices of the qubits to apply the analog evolution to. Defaults None for
            global application.
        :type index: Optional[List[int]]
        :param solver_options: Keyword arguments passed directly to `tc.timeevol.ode_evolve`
        :type solver_options: Dict[str, Any]
        """
        # Create and store the analog block information
        time = backend.convert_to_tensor(time, dtype=rdtypestr)
        time = backend.reshape(time, [-1])
        if backend.shape_tuple(time)[0] == 1:
            time = backend.stack([0.0, time[0]])  # type: ignore
        elif backend.shape_tuple(time)[0] > 2:
            raise ValueError(
                "Time must be a scalar or a two elements array for the starting and end points."
            )
        combined_solver_options = self._solver_options.copy()
        combined_solver_options.update(solver_options)
        block = AnalogBlock(
            hamiltonian_func=hamiltonian,
            time=time,  # type: ignore
            index=index,
            solver_options=combined_solver_options,
        )
        self.analog_blocks.append(block)

        # After adding an analog block, we start a new digital block.
        self.digital_circuits.append(Circuit(self.num_qubits, inputs=self.inputs))
        self._effective_circuit = None
        return self  # Allow for chaining

    def __getattr__(self, name: str) -> Any:
        """
        Metaprogramming to forward gate calls to the current digital circuit.
        This enables syntax like `analog_circuit.h(0)`.
        """
        gate_method = getattr(self.current_digital_circuit, name, None)

        if gate_method and callable(gate_method) and name.lower() in defined_gates:

            def wrapper(*args, **kwargs):  # type: ignore
                gate_method(*args, **kwargs)
                self._effective_circuit = None
                return self

            return wrapper
        else:
            raise AttributeError(
                f"'{type(self).__name__}' object or its underlying '{type(self.current_digital_circuit).__name__}' "
                f"object has no attribute '{name}'."
            )

    def state(self, form: str = "default") -> Tensor:
        """
        Executes the full digital-analog sequence.

        :return: The final state vector after the full evolution
        :rtype: Tensor
        """
        # Propagate the state through the alternating circuit blocks
        for i, analog_block in enumerate(self.analog_blocks):
            # 1. Apply Digital Block i
            digital_c = self.digital_circuits[i]
            if i > 0:
                digital_c.replace_inputs(psi)  # type: ignore
            psi = digital_c.wavefunction()

            if analog_block.index is None:
                psi = ode_evol_global(  # type: ignore
                    hamiltonian=analog_block.hamiltonian_func,
                    initial_state=psi,
                    times=analog_block.time,
                    **analog_block.solver_options,
                )
            else:
                psi = ode_evol_local(  # type: ignore
                    hamiltonian=analog_block.hamiltonian_func,
                    initial_state=psi,
                    times=analog_block.time,
                    index=analog_block.index,
                    **analog_block.solver_options,
                )
            psi = psi[-1]
            # TODO(@refraction-ray): support more time evol methods

        # 3. Apply the final digital circuit
        if self.analog_blocks:
            self.digital_circuits[-1].replace_inputs(psi)
            psi = self.digital_circuits[-1].wavefunction()
        else:
            psi = self.digital_circuits[-1].wavefunction()
        self._effective_circuit = Circuit(self.num_qubits, inputs=psi)
        if form == "default":
            shape = [-1]
        elif form == "ket":
            shape = [-1, 1]
        elif form == "bra":  # no conj here
            shape = [1, -1]
        return backend.reshape(psi, shape=shape)

    wavefunction = state

    def expectation(
        self,
        *ops: Tuple[tn.Node, List[int]],
        reuse: bool = True,
        enable_lightcone: bool = False,
        nmc: int = 1000,
        **kws: Any,
    ) -> Tensor:
        """
        Compute expectation(s) of local operators.

        :param ops: Pairs of `(operator_node, [sites])` specifying where each operator acts.
        :type ops: Tuple[tn.Node, List[int]]
        :param reuse: If True, then the wavefunction tensor is cached for further expectation evaluation,
            defaults to be true.
        :type reuse: bool, optional
        :param enable_lightcone: whether enable light cone simplification, defaults to False
        :type enable_lightcone: bool, optional
        :param nmc: repetition time for Monte Carlo sampling for noisfy calculation, defaults to 1000
        :type nmc: int, optional
        :return: Tensor with one element
        :rtype: Tensor
        """
        return self.effective_circuit.expectation(
            *ops,
            reuse=reuse,
            enable_lightcone=enable_lightcone,
            noise_conf=None,
            nmc=nmc,
            **kws,
        )

    def measure_jit(
        self, *index: int, with_prob: bool = False, status: Optional[Tensor] = None
    ) -> Tuple[Tensor, Tensor]:
        """
        Take measurement on the given site indices (computational basis).
        This method is jittable!

        :param index: Measure on which site (wire) index.
        :type index: int
        :param with_prob: If true, theoretical probability is also returned.
        :type with_prob: bool, optional
        :param status: external randomness, with shape [index], defaults to None
        :type status: Optional[Tensor]
        :return: The sample output and probability (optional) of the quantum line.
        :rtype: Tuple[Tensor, Tensor]
        """
        return self.effective_circuit.measure_jit(
            *index, with_prob=with_prob, status=status
        )

    measure = measure_jit

    def amplitude(self, l: Union[str, Tensor]) -> Tensor:
        r"""
        Return the amplitude for a given bitstring `l`.

        For state simulators, this computes :math:`\langle l \vert \psi \rangle`.

        :param l: Bitstring in base-`d` using `0-9A-Z`.
        :type l: Union[str, Tensor]
        :return: Complex amplitude.
        :rtype: Tensor
        """
        return self.effective_circuit.amplitude(l)

    def probability(self) -> Tensor:
        """
        Get the length-`2^n` probability vector over the computational basis.

        :return: Probability vector of shape `[dim^n]`.
        :rtype: Tensor
        """
        return self.effective_circuit.probability()

    def expectation_ps(
        self,
        x: Optional[Sequence[int]] = None,
        y: Optional[Sequence[int]] = None,
        z: Optional[Sequence[int]] = None,
        ps: Optional[Sequence[int]] = None,
        reuse: bool = True,
        noise_conf: Optional[Any] = None,
        nmc: int = 1000,
        status: Optional[Tensor] = None,
        **kws: Any,
    ) -> Tensor:
        """
        Shortcut for Pauli string expectation.
        x, y, z list are for X, Y, Z positions

        :Example:

        >>> c = tc.Circuit(2)
        >>> c.X(0)
        >>> c.H(1)
        >>> c.expectation_ps(x=[1], z=[0])
        array(-0.99999994+0.j, dtype=complex64)

        :param x: sites to apply X gate, defaults to None
        :type x: Optional[Sequence[int]], optional
        :param y: sites to apply Y gate, defaults to None
        :type y: Optional[Sequence[int]], optional
        :param z: sites to apply Z gate, defaults to None
        :type z: Optional[Sequence[int]], optional
        :param ps: or one can apply a ps structures instead of ``x``, ``y``, ``z``,
            e.g. [0, 1, 3, 0, 2, 2] for X_1Z_2Y_4Y_5
            defaults to None, ``ps`` can overwrite ``x``, ``y`` and ``z``
        :type ps: Optional[Sequence[int]], optional
        :param reuse: whether to cache and reuse the wavefunction, defaults to True
        :type reuse: bool, optional
        :param noise_conf: Noise Configuration, defaults to None
        :type noise_conf: Optional[NoiseConf], optional
        :param nmc: repetition time for Monte Carlo sampling for noisfy calculation, defaults to 1000
        :type nmc: int, optional
        :param status: external randomness given by tensor uniformly from [0, 1], defaults to None,
            used for noisfy circuit sampling
        :type status: Optional[Tensor], optional
        :return: Expectation value
        :rtype: Tensor
        """
        return self.effective_circuit.expectation_ps(
            x=x,
            y=y,
            z=z,
            ps=ps,
            reuse=reuse,
            noise_conf=noise_conf,
            nmc=nmc,
            status=status,
            **kws,
        )

    @partial(arg_alias, alias_dict={"format": ["format_"]})
    def sample(
        self,
        batch: Optional[int] = None,
        allow_state: bool = False,
        readout_error: Optional[Sequence[Any]] = None,
        format: Optional[str] = None,
        random_generator: Optional[Any] = None,
        status: Optional[Tensor] = None,
        jittable: bool = True,
    ) -> Any:
        r"""
        Batched sampling from the circuit or final state.

        :param batch: Number of samples. If `None`, returns a single draw.
        :type batch: Optional[int]
        :param allow_state: If `True`, sample from the final state (when memory allows). Prefer `True` for speed.
        :type allow_state: bool
        :param readout_error: Optional readout error model.
        :type readout_error: Optional[Sequence[Any]]
        :param format: Output format. See :py:meth:`tensorcircuit.quantum.measurement_results`.
        :type format: Optional[str]
        :param random_generator: random generator,  defaults to None
        :type random_generator: Optional[Any], optional
        :param status: external randomness given by tensor uniformly from [0, 1],
            if set, can overwrite random_generator, shape [batch] for `allow_state=True`
            and shape [batch, nqudits] for `allow_state=False` using perfect sampling implementation
        :type status: Optional[Tensor]
        :param jittable: when converting to count, whether keep the full size. if false, may be conflict
            external jit, if true, may fail for large scale system with actual limited count results
        :type jittable: bool, defaults true
        :return: List (if batch) of tuple (binary configuration tensor and corresponding probability)
            if the format is None, and consistent with format when given
        :rtype: Any
        """
        return self.effective_circuit.sample(
            batch=batch,
            allow_state=allow_state,
            readout_error=readout_error,
            format=format,
            random_generator=random_generator,
            status=status,
            jittable=jittable,
        )

    def __repr__(self) -> str:
        s = f"AnalogCircuit(n={self.num_qubits}):\n"
        s += "=" * 40 + "\n"

        num_stages = len(self.analog_blocks) + 1

        for i in range(num_stages):
            # Print digital part
            s += f"--- Digital Block {i} ---\n"

            # Print analog part (if it exists)
            if i < len(self.analog_blocks):
                block = self.analog_blocks[i]
                s += f"--- Analog Block {i} (T={block.time}) ---\n"
                s += f" H(t) function: '{block.hamiltonian_func.__name__}'\n"

        s += "=" * 40
        return s



from . import ansatz
from . import blocks
from . import chems
from . import dataset
from . import graphs
from . import measurements
from . import conversions
from . import lattice
from . import hamiltonians

costfunctions = measurements
